# Database
DATABASE_HOST="localhost"
DATABASE_USER="magu"
DATABASE_PASSWORD="magu"
DATABASE_PORT=3306
DATABASE_NAME="magu"

# Server
SERVER_HOST="localhost"
SERVER_PORT=8000
SERVER_ADDRESS = f"http://{SERVER_HOST}:{SERVER_PORT}"

# Authentication
SECRET_KEY="secret_key_123"



from magu.core import router
import re

def request_mapping(uri: str = "/"):
    def wrapper(cls):
        cls._uri = uri
        return cls
    return wrapper

def parse_route(route: str):
    param_regex = r"{([a-zA-Z_][a-zA-Z0-9_]*)}"
    params = re.findall(param_regex, route)

    pattern = re.sub(param_regex, lambda m: f"(?P<{m.group(1)}>[^/]+)", route)
    pattern = "^" + pattern + "$"

    return re.compile(pattern), params


class PathVariable:
    def __init__(self):
        self.name = None
    
    def __set_name__(self, owner, name: str):
        self.name = name
    
    def __get__(self, instance, owner):
        if instance is None:
            return self
        return instance.__dict__.get(self.name)

    def __set__(self, instance, value):
        instance.__dict__[self.name] = value


def get_mapping(uri: str = "/"):
        def wrapper(func):
            func._http_method = 'GET'

            pattern = r"({(.*?)})"
            uri_initial = re.sub(pattern, "{ID_PLACEHOLDER}", uri)
            func._method_uri = uri_initial

            return func
        return wrapper
    
def post_mapping(uri: str = "/"):
    def wrapper(func):
        func._http_method = 'POST'

        pattern = r"({(.*?)})"
        uri_initial = re.sub(pattern, "{ID_PLACEHOLDER}", uri)
        func._method_uri = uri_initial

        return func
    return wrapper
    
def put_mapping(uri: str = "/"):
    def wrapper(func):
        func._http_method = 'PUT'

        pattern = r"({(.*?)})"
        uri_initial = re.sub(pattern, "{ID_PLACEHOLDER}", uri)
        func._method_uri = uri_initial
        
        return func
    return wrapper
    
def patch_mapping(uri: str = "/"):
    def wrapper(func):
        func._http_method = 'PATCH'

        pattern = r"({(.*?)})"
        uri_initial = re.sub(pattern, "{ID_PLACEHOLDER}", uri)
        func._method_uri = uri_initial

        return func
    return wrapper
    
def delete_mapping(uri: str = "/"):
    def wrapper(func):
        func._http_method = 'DELETE'

        pattern = r"({(.*?)})"
        uri_initial = re.sub(pattern, "{ID_PLACEHOLDER}", uri)
        func._method_uri = uri_initial
        
        return func
    return wrapper

class Controller:
    def __init__(self):
        self.uri = self._uri
        router.get_routes(self.__class__)

    def __str__(self):
        return f"Controller for {self.uri}"



def table(name: str, user: bool = False):
    def wrapper(cls):
        cls._table = name
        cls._user = user
        return cls
    return wrapper

class Model:
    def __init__(self):
        try:
            self._table
        except Exception:
            print(f"[Model] {self.__class__.__name__} has no table name defined at @table()")

    def __call__(self):
        return self.__dict__

# Descriptor para marcar atributos (como annotations, em Java)
class Column:
    def __init__(self,
                type,
                name: str = None,
                nullable: bool = True,
                unique: bool = False,
                id: bool = False,
                many_to_one: bool = False):
       self.type = type
       self.name = name
       self.nullable = nullable
       self.unique = unique
       self.id = id
       self.many_to_one = many_to_one

    def __set_name__(self, owner, name: str):
        self.name = name if not self.name else self.name
    
    def __get__(self, instance, owner):
        if instance is None:
            return self
        return instance.__dict__.get(self.name)
    
    def __set__(self, instance, value):
        instance.__dict__[self.name] = value



from magu.database.database import MySQLDatabase as mysql
#from magu.database.database import MariaDBDatabase as mariadb
from magu.models.model import Model
from magu.models.model import Column

class MySQLRepository:
    def __init__(self, model: Model.__class__):
        self.model = model

    def save(self):
        table = self.model._table

        for attr in self.model.__dict__:
            print(attr, self.model.__dict__[attr])

        #with mysql.session() as session:
            ...#session.execute("INSERT INTO %s VALUEs ()", (table,))

class MariaDBRepository:
    def __init__(self, model: Model.__class__):
        self.model = model
        self.table = self.model._table
        self.columns = [
            name for name, attr in self.model.__dict__.items()
            if isinstance(attr, Column)
            if not attr.id
        ]
        self.columns_id = [
            name for name, attr in self.model.__dict__.items()
            if isinstance(attr, Column)
        ]

        for name, attr in self.model.__dict__.items():
            if isinstance(attr, Column) and attr.id:
                self.id = name

    def find_all(self):
        with mariadb.session() as session:
            session.execute(f"SELECT * FROM `{self.table}`")
            res = session.fetchall()
            for i in range(len(res)):
                res[i] = dict(zip(self.columns_id, res[i]))
            return res

    def find_by_id(self, id):
        with mariadb.session() as session:
            session.execute(f"SELECT * FROM `{self.table}` WHERE `{self.id}` = {id}")
            return dict(zip(self.columns_id, session.fetchone()))
        

    def save(self, entity: Model):
        values = [getattr(entity, col) for col in self.columns]

        placeholders = ",".join("%s" for _ in self.columns)

        query = f"INSERT INTO `{self.table}` ({",".join(f"`{value}`" for value in self.columns)}) VALUES ({placeholders});"
        fetch = f"SELECT * FROM `{self.table}` WHERE `{self.id}` = %s"
        with mariadb.session() as session:
            session.execute(query, values)
            
            last_id = session.lastrowid
            entity.id = last_id

            session.execute(fetch, (last_id,))
            res = dict(zip(self.columns_id, session.fetchone()))
            return res

    def update(self, id, new_model):
        # pega apenas colunas definidas no model (exceto o id)
        columns = [
            name for name, attr in self.model.__dict__.items()
            if isinstance(attr, Column) and not attr.id
        ]

        # pega os novos valores que o usuário enviou
        values = [getattr(new_model, col) for col in columns]

        # gera: "name = %s, age = %s, ..."
        set_clause = ", ".join(f"`{col}` = %s" for col in columns)

        query = f"""
            UPDATE `{self.table}`
            SET {set_clause}
            WHERE `{self.id}` = %s
        """

        with mariadb.session() as session:
            session.execute(query, (*values, id))  # executa UPDATE

            # busca o registro atualizado
            session.execute(
                f"SELECT * FROM `{self.table}` WHERE `{self.id}` = %s",
                (id,)
            )

            return dict(zip(self.columns_id, session.fetchone()))
